#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from ..data import check_input
from ..transform.filter import filter_target_decoy
from ..transform.util import get_available_keys
from .util import __get_filename

from typing import Optional
from typing import Dict
from typing import Any
from typing import List


def __csms_to_xifdr(
    csms: List[Dict[str, Any]],
    filename: Optional[str],
) -> pd.DataFrame:
    r"""Exports crosslink-spectrum-matches to xiFDR format.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.
    filename : str, or None
        If not None, the data will be written to a file with the specified filename.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame in xiFDR format.

    Notes
    -----
    This function should not be called directly, it is called from ``to_xifdr()``.
    """
    run = list()
    scan = list()
    peptide1 = list()
    peptide2 = list()
    peptidelink1 = list()
    peptidelink2 = list()
    isdecoy1 = list()
    isdecoy2 = list()
    precursorcharge = list()
    accession1 = list()
    accession2 = list()
    peptideposition1 = list()
    peptideposition2 = list()
    score = list()
    for csm in csms:
        run.append(csm["spectrum_file"])
        scan.append(csm["scan_nr"])
        peptide1.append(csm["alpha_peptide"])
        peptide2.append(csm["beta_peptide"])
        peptidelink1.append(csm["alpha_peptide_crosslink_position"])
        peptidelink2.append(csm["beta_peptide_crosslink_position"])
        isdecoy1.append("true" if csm["alpha_decoy"] else "false")
        isdecoy2.append("true" if csm["beta_decoy"] else "false")
        precursorcharge.append(csm["charge"])
        accession1.append(";".join(csm["alpha_proteins"]))
        accession2.append(";".join(csm["beta_proteins"]))
        peptideposition1.append(
            ";".join([str(pos) for pos in csm["alpha_proteins_peptide_positions"]])
        )
        peptideposition2.append(
            ";".join([str(pos) for pos in csm["beta_proteins_peptide_positions"]])
        )
        score.append(csm["score"])
    xifdr_df = pd.DataFrame(
        {
            "run": run,
            "scan": scan,
            "peptide1": peptide1,
            "peptide2": peptide2,
            "peptide link 1": peptidelink1,
            "peptide link 2": peptidelink2,
            "is decoy 1": isdecoy1,
            "is decoy 2": isdecoy2,
            "precursor charge": precursorcharge,
            "accession1": accession1,
            "accession2": accession2,
            "peptide position 1": peptideposition1,
            "peptide position 2": peptideposition2,
            "score": score,
        }
    )
    if filename is not None:
        xifdr_df.to_csv(__get_filename(filename, "csv"), index=False)
    return xifdr_df


def to_xifdr(
    csms: List[Dict[str, Any]],
    filename: Optional[str],
) -> pd.DataFrame:
    r"""Exports a list of crosslink-spectrum-matches to xiFDR format.

    Exports a list of crosslinks to xiFDR format. The tool xiFDR is accessible
    via the link
    `rappsilberlab.org/software/xifdr <https://www.rappsilberlab.org/software/xifdr/>`_.
    Requires that ``alpha_proteins``, ``beta_proteins``, ``alpha_proteins_peptide_positions``,
    ``beta_proteins_peptide_positions``, ``alpha_decoy``, ``beta_decoy``, ``charge`` and ``score``
    fields are set for all crosslink-spectrum-matches.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.
    filename : str, or None
        If not None, the exported data will be written to a file with the specified filename.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing crosslink-spectrum-matches in xiFDR format.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If 'csms' parameter contains elements of mixed data type.
    ValueError
        If the provided 'csms' parameter contains no elements.
    RuntimeError
        If not all of the required information is present in the input data.

    Examples
    --------
    >>> from pyXLMS.exporter import to_xifdr
    >>> from pyXLMS.parser import read
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> csms = pr["crosslink-spectrum-matches"]
    >>> to_xifdr(csms, filename="msannika_xiFDR.csv")
                                           run   scan          peptide1  ... peptide position 1  peptide position 2   score
    0    XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw   2257            GQKNSR  ...                777                 777  119.83
    1    XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw   2448            GQKNSR  ...                777                 693   13.91
    2    XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw   2561             SDKNR  ...                864                 864  114.43
    3    XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw   2719            DKQSGK  ...                676                 676  200.98
    4    XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw   2792            DKQSGK  ...                676                  45   94.47
    ..                                     ...    ...               ...  ...                ...                 ...     ...
    821  XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw  23297     MDGTEELLVKLNR  ...                387                 387  286.05
    822  XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw  23454  KIECFDSVEISGVEDR  ...                575                 682  376.15
    823  XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw  23581    SSFEKNPIDFLEAK  ...               1176                1176  412.44
    824  XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw  23683    SSFEKNPIDFLEAK  ...               1176                1176  437.10
    825  XLpeplib_Beveridge_QEx-HFX_DSS_R1.raw  27087    MEDESKLHKFKDFK  ...                 99                1176   15.89
    [826 rows x 14 columns]

    >>> from pyXLMS.exporter import to_xifdr
    >>> from pyXLMS.parser import read
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> csms = pr["crosslink-spectrum-matches"]
    >>> df = to_xifdr(csms, filename=None)
    """
    _ok = check_input(csms, "csms", list, dict)
    _ok = check_input(filename, "filename", str) if filename is not None else True
    if len(csms) == 0:
        raise ValueError("Provided crosslink-spectrum-matches contain no elements!")
    if "data_type" not in csms[0] or csms[0]["data_type"] != "crosslink-spectrum-match":
        raise TypeError(
            "Unsupported data type for input csms! Parameter csms has to be a list of crosslink-spectrum-matches!"
        )
    available_keys = get_available_keys(csms)
    if (
        not available_keys["alpha_proteins"]
        or not available_keys["beta_proteins"]
        or not available_keys["alpha_proteins_crosslink_positions"]
        or not available_keys["beta_proteins_crosslink_positions"]
        or not available_keys["alpha_decoy"]
        or not available_keys["beta_decoy"]
        or not available_keys["charge"]
        or not available_keys["score"]
        or len(filter_target_decoy(csms)["Target-Decoy"]) == 0
    ):
        raise RuntimeError(
            "Can't export to xiFDR because not all necessary information is available!"
        )
    return __csms_to_xifdr(csms, filename)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

__all__ = [
    "read_xi",
    "read_mzid",
    "read_plink",
    "read_scout",
    "read_xlinkx",
    "read_custom",
    "read_merox",
    "read_msannika",
    "read_maxquant",
    "read_maxlynx",
    "detect_xi_filetype",
    "parse_peptide",
    "parse_modifications_from_xi_sequence",
    "parse_scan_nr_from_mzid",
    "parse_scan_nr_from_plink",
    "parse_spectrum_file_from_plink",
    "detect_plink_filetype",
    "detect_scout_filetype",
    "parse_modifications_from_scout_sequence",
    "pyxlms_modification_str_parser",
    "parse_modifications_from_maxquant_sequence",
    "read",
]

# READERS
from .parser_xldbse_xi import read_xi
from .parser_xldbse_mzid import read_mzid
from .parser_xldbse_plink import read_plink
from .parser_xldbse_scout import read_scout
from .parser_xldbse_xlinkx import read_xlinkx
from .parser_xldbse_custom import read_custom
from .parser_xldbse_merox import read_merox
from .parser_xldbse_msannika import read_msannika
from .parser_xldbse_maxquant import read_maxquant
from .parser_xldbse_maxquant import read_maxlynx

# UTILITY
from .parser_xldbse_xi import detect_xi_filetype
from .parser_xldbse_xi import parse_peptide
from .parser_xldbse_xi import parse_modifications_from_xi_sequence
from .parser_xldbse_mzid import parse_scan_nr_from_mzid
from .parser_xldbse_plink import parse_scan_nr_from_plink
from .parser_xldbse_plink import parse_spectrum_file_from_plink
from .parser_xldbse_plink import detect_plink_filetype
from .parser_xldbse_scout import detect_scout_filetype
from .parser_xldbse_scout import parse_modifications_from_scout_sequence
from .parser_xldbse_custom import pyxlms_modification_str_parser
from .parser_xldbse_maxquant import parse_modifications_from_maxquant_sequence

from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import List

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def read(
    files: str | List[str] | BinaryIO,
    engine: Literal[
        "Custom",
        "MaxQuant",
        "MaxLynx",
        "MeroX",
        "MS Annika",
        "mzIdentML",
        "pLink",
        "Scout",
        "xiSearch/xiFDR",
        "XlinkX",
    ],
    crosslinker: str,
    parse_modifications: bool = True,
    ignore_errors: bool = False,
    verbose: Literal[0, 1, 2] = 1,
    **kwargs,
) -> Dict[str, Any]:
    r"""Read a crosslink result file.

    Reads a crosslink or crosslink-spectrum-match result file from any of the supported crosslink search engines or formats.
    Currently supports results files from MaxLynx/MaxQuant, MeroX, MS Annika, pLink 2 and pLink 3, Scout, xiSearch and xiFDR,
    XlinkX, and the mzIdentML format. Additionally supports parsing from custom ``.csv`` files in pyXLMS format, see more
    about the custom format in ``parser.read_custom()`` and in here:
    `docs <https://github.com/hgb-bin-proteomics/pyXLMS/blob/master/docs/format.md>`_.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the result file(s) or a file-like object/stream.
    engine : "Custom", "MaxQuant", "MaxLynx", "MeroX", "MS Annika", "mzIdentML", "pLink", "Scout", "xiSearch/xiFDR", or "XlinkX"
        Crosslink search engine or format of the result file.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter for every parser. Defaults are selected
        for every parser if 'modifications' is not passed via ``**kwargs``.
    ignore_errors : bool, default = False
        Ignore errors when mapping modifications. Used in ``parser.read_xi()`` and ``parser.read_xlinkx()``.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.
    **kwargs
        Any additional parameters will be passed to the specific parsers.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    ValueError
        If the value entered for parameter ``engine`` is not supported.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> csms_from_xiSearch = read(
    ...     "data/xi/r1_Xi1.7.6.7.csv", engine="xiSearch/xiFDR", crosslinker="DSS"
    ... )

    >>> from pyXLMS.parser import read
    >>> csms_from_MaxQuant = read(
    ...     "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    ... )
    """
    supported = [
        "Custom",
        "MaxQuant",
        "MaxLynx",
        "MeroX",
        "MS Annika",
        "mzIdentML",
        "pLink",
        "Scout",
        "xiSearch/xiFDR",
        "XlinkX",
    ]
    ff = engine.lower().strip()

    if ff in ["custom", "pyxlms"]:
        return read_custom(files, parse_modifications=parse_modifications, **kwargs)
    if ff in ["maxquant", "max quant"]:
        return read_maxquant(
            files,
            crosslinker=crosslinker,
            parse_modifications=parse_modifications,
            **kwargs,
        )
    if ff in ["maxlynx", "max lynx"]:
        return read_maxlynx(
            files,
            crosslinker=crosslinker,
            parse_modifications=parse_modifications,
            **kwargs,
        )
    if ff in ["merox", "stavrox"]:
        return read_merox(
            files,
            crosslinker=crosslinker,
            parse_modifications=parse_modifications,
            **kwargs,
        )
    if ff in ["ms annika", "msannika"]:
        return read_msannika(
            files, parse_modifications=parse_modifications, verbose=verbose, **kwargs
        )
    if ff in ["mzidentml", "mzid"]:
        return read_mzid(files, verbose=verbose, **kwargs)
    if ff in ["plink", "plink2", "plink3", "plink 2", "plink 3"]:
        return read_plink(
            files, parse_modifications=parse_modifications, verbose=verbose, **kwargs
        )
    if ff in ["scout"]:
        return read_scout(
            files,
            crosslinker=crosslinker,
            parse_modifications=parse_modifications,
            verbose=verbose,
            **kwargs,
        )
    if ff in ["xisearch/xifdr", "xisearch", "xifdr", "xi search", "xi fdr", "xi"]:
        return read_xi(
            files,
            parse_modifications=parse_modifications,
            ignore_errors=ignore_errors,
            verbose=verbose,
            **kwargs,
        )
    if ff in ["xlinkx", "x link x"]:
        return read_xlinkx(
            files,
            parse_modifications=parse_modifications,
            ignore_errors=ignore_errors,
            verbose=verbose,
            **kwargs,
        )

    err_str = (
        f"{engine} is not a supported crosslink search engine or format! Valid options are:\n"
        + ", ".join(supported)
    )
    raise ValueError(err_str)

    return {"err": err_str}



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

__all__ = [
    "constants",
    "data",
    "exporter",
    "parser",
    "pipelines",
    "transform",
    "plotting",
]
__version__ = "1.7.3"
__author__ = "Micha Johannes Birklbauer"

from . import constants
from . import data
from . import exporter
from . import parser
from . import pipelines
from . import transform
from . import plotting



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com


def test1():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_target_decoy

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    target_and_decoys = filter_target_decoy(result["crosslink-spectrum-matches"])
    assert len(target_and_decoys["Target-Target"]) == 786
    assert len(target_and_decoys["Target-Decoy"]) == 39
    assert len(target_and_decoys["Decoy-Decoy"]) == 1


def test2():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_target_decoy

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    target_and_decoys = filter_target_decoy(result["crosslinks"])
    assert len(target_and_decoys["Target-Target"]) == 265
    assert len(target_and_decoys["Target-Decoy"]) == 0
    assert len(target_and_decoys["Decoy-Decoy"]) == 35


def test3():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_proteins

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    proteins_csms = filter_proteins(result["crosslink-spectrum-matches"], ["Cas9"])
    assert proteins_csms["Proteins"] == ["Cas9"]
    assert len(proteins_csms["Both"]) == 798
    assert len(proteins_csms["One"]) == 23


def test4():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_proteins

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    proteins_xls = filter_proteins(result["crosslinks"], ["Cas9"])
    assert proteins_xls["Proteins"] == ["Cas9"]
    assert len(proteins_xls["Both"]) == 274
    assert len(proteins_xls["One"]) == 21


def test5():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_crosslink_type

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslink_type_filtered_csms = filter_crosslink_type(
        result["crosslink-spectrum-matches"]
    )
    assert len(crosslink_type_filtered_csms["Intra"]) == 803
    assert len(crosslink_type_filtered_csms["Inter"]) == 23


def test6():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_crosslink_type

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    crosslink_type_filtered_crosslinks = filter_crosslink_type(result["crosslinks"])
    assert len(crosslink_type_filtered_crosslinks["Intra"]) == 279
    assert len(crosslink_type_filtered_crosslinks["Inter"]) == 21


def test7():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_protein_distribution

    result = read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    proteins_csms = filter_protein_distribution(result["crosslink-spectrum-matches"])
    proteins_found = list(proteins_csms.keys())  # proteins found
    proteins = [
        "Cas9",
        "sp|MYG_HUMAN|",
        "sp|CAH1_HUMAN|",
        "sp|RETBP_HUMAN|",
        "sp|K1C15_SHEEP|",
    ]
    for p in proteins:
        assert p in proteins_found
    cas9 = len(proteins_csms["Cas9"])  # number of CSMs for protein Cas9
    assert cas9 == 728


def test8():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_peptide_pair_distribution

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    peptide_pairs = filter_peptide_pair_distribution(
        result["crosslink-spectrum-matches"]
    )
    peptide_pairs_found = list(peptide_pairs.keys())[:5]  # first 5 found peptide pairs
    peptide_pairs_should = [
        "GQKNSR-GQKNSR",
        "GQKNSR-GSQKDR",
        "SDKNR-SDKNR",
        "DKQSGK-DKQSGK",
        "DKQSGK-HSIKK",
    ]
    for p in peptide_pairs_should:
        assert p in peptide_pairs_found
    MTNFDKNLPNEK_SKLVSDFR = len(
        peptide_pairs["MTNFDKNLPNEK-SKLVSDFR"]
    )  # number of CSMs for peptide pair MTNFDKNLPNEK-SKLVSDFR
    assert MTNFDKNLPNEK_SKLVSDFR == 21


def test9():
    from pyXLMS.parser import read
    from pyXLMS.transform import filter_peptide_pair_distribution
    from pyXLMS.transform import aggregate

    result = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    peptide_pairs = filter_peptide_pair_distribution(
        result["crosslink-spectrum-matches"]
    )
    assert len(peptide_pairs) == len(aggregate(result["crosslink-spectrum-matches"]))



#!/usr/bin/env python3

# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd
from ..data import check_input
from .util import modifications_to_str

from typing import Optional
from typing import List
from typing import Dict
from typing import Any


def __cc(input_list: Optional[List[Any]], sep: str = ";") -> str | None:
    r"""Concatenates list elements to a string using the defined seperator.

    Parameters
    ----------
    input_list : list, or None
        The list to concatenate.
    sep : str, default = ";"
        The seperator to use for concatentation.

    Returns
    -------
    str, or None
        The concatenated string of the list (or ``None`` if no list was provided).
    """
    s = ""
    if input_list is None:
        return None
    for i in input_list:
        s += str(i).strip() + sep
    return s.rstrip(sep)


def __crosslinks_to_dataframe(data: List[Dict[str, Any]]) -> pd.DataFrame:
    r"""Returns a pandas DataFrame of the given crosslinks.

    Parameters
    ----------
    data : list
        A list of crosslinks as created by ``data.create_crosslink()``.

    Returns
    -------
    pandas.DataFrame
        The pandas DataFrame created from the list of input crosslinks.

    Raises
    ------
    TypeError
        If the list does not contain crosslinks.
    ValueError
        If the list does not contain any objects.

    Notes
    -----
    This function should not be called directly, it is called from ``to_dataframe()``.
    """
    ## columns
    completeness = list()
    alpha_peptide = list()
    alpha_peptide_crosslink_position = list()
    alpha_proteins = list()
    alpha_proteins_crosslink_positions = list()
    alpha_decoy = list()
    beta_peptide = list()
    beta_peptide_crosslink_position = list()
    beta_proteins = list()
    beta_proteins_crosslink_positions = list()
    beta_decoy = list()
    crosslink_type = list()
    score = list()
    ## assign values
    for crosslink in data:
        completeness.append(crosslink["completeness"])
        alpha_peptide.append(crosslink["alpha_peptide"])
        alpha_peptide_crosslink_position.append(
            crosslink["alpha_peptide_crosslink_position"]
        )
        alpha_proteins.append(__cc(crosslink["alpha_proteins"]))
        alpha_proteins_crosslink_positions.append(
            __cc(crosslink["alpha_proteins_crosslink_positions"])
        )
        alpha_decoy.append(crosslink["alpha_decoy"])
        beta_peptide.append(crosslink["beta_peptide"])
        beta_peptide_crosslink_position.append(
            crosslink["beta_peptide_crosslink_position"]
        )
        beta_proteins.append(__cc(crosslink["beta_proteins"]))
        beta_proteins_crosslink_positions.append(
            __cc(crosslink["beta_proteins_crosslink_positions"])
        )
        beta_decoy.append(crosslink["beta_decoy"])
        crosslink_type.append(crosslink["crosslink_type"])
        score.append(crosslink["score"])
    return pd.DataFrame(
        {
            "Completeness": completeness,
            "Alpha Peptide": alpha_peptide,
            "Alpha Peptide Crosslink Position": alpha_peptide_crosslink_position,
            "Alpha Proteins": alpha_proteins,
            "Alpha Proteins Crosslink Positions": alpha_proteins_crosslink_positions,
            "Alpha Decoy": alpha_decoy,
            "Beta Peptide": beta_peptide,
            "Beta Peptide Crosslink Position": beta_peptide_crosslink_position,
            "Beta Proteins": beta_proteins,
            "Beta Proteins Crosslink Positions": beta_proteins_crosslink_positions,
            "Beta Decoy": beta_decoy,
            "Crosslink Type": crosslink_type,
            "Crosslink Score": score,
        }
    )


def __csms_to_dataframe(data: List[Dict[str, Any]]) -> pd.DataFrame:
    r"""Returns a pandas DataFrame of the given crosslink-spectrum-matches.

    Parameters
    ----------
    data : list
        A list of crosslink-spectrum-matches as created by ``data.create_csm()``.

    Returns
    -------
    pandas.DataFrame
        The pandas DataFrame created from the list of input crosslink-spectrum-matches.

    Raises
    ------
    TypeError
        If the list does not contain crosslink-spectrum-matches.
    ValueError
        If the list does not contain any objects.

    Notes
    -----
    This function should not be called directly, it is called from ``to_dataframe()``.
    """
    ## columns
    completeness = list()
    alpha_peptide = list()
    alpha_modifications = list()
    alpha_peptide_crosslink_position = list()
    alpha_proteins = list()
    alpha_proteins_crosslink_positions = list()
    alpha_proteins_peptide_positions = list()
    alpha_score = list()
    alpha_decoy = list()
    beta_peptide = list()
    beta_modifications = list()
    beta_peptide_crosslink_position = list()
    beta_proteins = list()
    beta_proteins_crosslink_positions = list()
    beta_proteins_peptide_positions = list()
    beta_score = list()
    beta_decoy = list()
    crosslink_type = list()
    score = list()
    spectrum_file = list()
    scan_nr = list()
    charge = list()
    retention_time = list()
    ion_mobility = list()
    ## assign values
    for csm in data:
        completeness.append(csm["completeness"])
        alpha_peptide.append(csm["alpha_peptide"])
        alpha_modifications.append(modifications_to_str(csm["alpha_modifications"]))
        alpha_peptide_crosslink_position.append(csm["alpha_peptide_crosslink_position"])
        alpha_proteins.append(__cc(csm["alpha_proteins"]))
        alpha_proteins_crosslink_positions.append(
            __cc(csm["alpha_proteins_crosslink_positions"])
        )
        alpha_proteins_peptide_positions.append(
            __cc(csm["alpha_proteins_peptide_positions"])
        )
        alpha_score.append(csm["alpha_score"])
        alpha_decoy.append(csm["alpha_decoy"])
        beta_peptide.append(csm["beta_peptide"])
        beta_modifications.append(modifications_to_str(csm["beta_modifications"]))
        beta_peptide_crosslink_position.append(csm["beta_peptide_crosslink_position"])
        beta_proteins.append(__cc(csm["beta_proteins"]))
        beta_proteins_crosslink_positions.append(
            __cc(csm["beta_proteins_crosslink_positions"])
        )
        beta_proteins_peptide_positions.append(
            __cc(csm["beta_proteins_peptide_positions"])
        )
        beta_score.append(csm["beta_score"])
        beta_decoy.append(csm["beta_decoy"])
        crosslink_type.append(csm["crosslink_type"])
        score.append(csm["score"])
        spectrum_file.append(csm["spectrum_file"])
        scan_nr.append(csm["scan_nr"])
        charge.append(csm["charge"])
        retention_time.append(csm["retention_time"])
        ion_mobility.append(csm["ion_mobility"])
    return pd.DataFrame(
        {
            "Completeness": completeness,
            "Alpha Peptide": alpha_peptide,
            "Alpha Peptide Modifications": alpha_modifications,
            "Alpha Peptide Crosslink Position": alpha_peptide_crosslink_position,
            "Alpha Proteins": alpha_proteins,
            "Alpha Proteins Crosslink Positions": alpha_proteins_crosslink_positions,
            "Alpha Proteins Peptide Positions": alpha_proteins_peptide_positions,
            "Alpha Score": alpha_score,
            "Alpha Decoy": alpha_decoy,
            "Beta Peptide": beta_peptide,
            "Beta Peptide Modifications": beta_modifications,
            "Beta Peptide Crosslink Position": beta_peptide_crosslink_position,
            "Beta Proteins": beta_proteins,
            "Beta Proteins Crosslink Positions": beta_proteins_crosslink_positions,
            "Beta Proteins Peptide Positions": beta_proteins_peptide_positions,
            "Beta Score": beta_score,
            "Beta Decoy": beta_decoy,
            "Crosslink Type": crosslink_type,
            "CSM Score": score,
            "Spectrum File": spectrum_file,
            "Scan Nr": scan_nr,
            "Precursor Charge": charge,
            "Retention Time": retention_time,
            "Ion Mobility": ion_mobility,
        }
    )


def to_dataframe(data: List[Dict[str, Any]]) -> pd.DataFrame:
    r"""Returns a pandas DataFrame of the given crosslinks or crosslink-spectrum-matches.

    Parameters
    ----------
    data : list
        A list of crosslinks or crosslink-spectrum-matches as created by ``data.create_crosslink()`` or ``data.create_csm()``.

    Returns
    -------
    pandas.DataFrame
        The pandas DataFrame created from the list of input crosslinks or crosslink-spectrum-matches.
        A full specification of the returned DataFrame can be found in the
        `docs <https://github.com/hgb-bin-proteomics/pyXLMS/blob/master/docs/format.md>`_.

    Raises
    ------
    TypeError
        If the list does not contain crosslinks or crosslink-spectrum-matches.
    ValueError
        If the list does not contain any objects.

    Examples
    --------
    >>> from pyXLMS.transform import to_dataframe
    >>> # assume that crosslinks is a list of crosslinks created by data.create_crosslink()
    >>> crosslink_dataframe = to_dataframe(crosslinks)
    >>> # assume csms is a list of crosslink-spectrum-matches created by data.create_csm()
    >>> csm_dataframe = to_dataframe(csms)
    """
    ## input checks
    check_input(data, "data", list, dict)
    ## function calls
    if len(data) > 0:
        if "data_type" in data[0] and data[0]["data_type"] == "crosslink":
            return __crosslinks_to_dataframe(data)
        elif (
            "data_type" in data[0]
            and data[0]["data_type"] == "crosslink-spectrum-match"
        ):
            return __csms_to_dataframe(data)
        else:
            raise TypeError("The given data object is not supported!")
    else:
        raise ValueError("Parameter data has to be at least of length one!")



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import warnings
import pandas as pd
from tqdm import tqdm

from ..data import check_input
from ..data import create_crosslink
from ..data import create_csm
from ..data import create_parser_result
from ..constants import XI_MODIFICATION_MAPPING
from .util import format_sequence
from .util import get_bool_from_value
from .util import __serialize_pandas_series

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def detect_xi_filetype(
    data: pd.DataFrame,
) -> Literal["xisearch", "xifdr_csms", "xifdr_crosslinks"]:
    r"""Detects the xi-related source (application) of the data.

    Detects whether the input data is originating from xiSearch or xiFDR, and if xiFDR which type of data is
    being read (crosslink-spectrum-matches or crosslinks).

    Parameters
    ----------
    data : pd.DataFrame
        The input data originating from xiSearch or xiFDR.

    Returns
    -------
    str
        "xisearch" if a xiSearch result file was read, "xifdr_csms" if CSMs from xiFDR were read,
        "xifdr_crosslinks" if crosslinks from xiFDR were read.

    Raises
    ------
    ValueError
        If the data source could not be determined.

    Examples
    --------
    >>> from pyXLMS.parser import detect_xi_filetype
    >>> import pandas as pd
    >>> df1 = pd.read_csv("data/xi/r1_Xi1.7.6.7.csv")
    >>> detect_xi_filetype(df1)
    'xisearch'

    >>> from pyXLMS.parser import detect_xi_filetype
    >>> import pandas as pd
    >>> df2 = pd.read_csv("data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv")
    >>> detect_xi_filetype(df2)
    'xifdr_csms'

    >>> from pyXLMS.parser import detect_xi_filetype
    >>> import pandas as pd
    >>> df3 = pd.read_csv("data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv")
    >>> detect_xi_filetype(df3)
    'xifdr_crosslinks'
    """
    ## check input
    _ok = check_input(data, "data", pd.DataFrame)

    col_names = data.columns.values.tolist()
    if "AllScore" in col_names:
        return "xisearch"
    if "LinkPos1" in col_names:
        return "xifdr_csms"
    if "ToSite" in col_names:
        return "xifdr_crosslinks"

    raise ValueError(
        "Could not infer data source, are you sure you read a xi result file?"
    )

    return "err"


def parse_peptide(sequence: str, term_char: str = ".") -> str:
    r"""Parses the peptide sequence from a sequence string including flanking amino acids.

    Parses the peptide sequence from a sequence string including flanking amino acids, for example ``"K.KKMoxKLS.S"``.
    The returned peptide sequence for this example would be ``"KKMoxKLS"``.

    Parameters
    ----------
    sequence : str
        The sequence string containing the peptide sequence and flanking amino acids.
    term_char : str (single character), default = "."
        The character used to denote N-terminal and C-terminal.

    Returns
    -------
    str
        The parsed peptide sequence without flanking amino acids.

    Raises
    ------
    RuntimeError
        If (one of) the peptide sequence(s) could not be parsed.

    Examples
    --------
    >>> from pyXLMS.parser import parse_peptide
    >>> parse_peptide("K.KKMoxKLS.S")
    'KKMoxKLS'

    >>> from pyXLMS.parser import parse_peptide
    >>> parse_peptide("-.CcmCcmPSR.T")
    'CcmCcmPSR'

    >>> from pyXLMS.parser import parse_peptide
    >>> parse_peptide("CCPSR")
    'CCPSR'
    """
    ## check input
    _ok = check_input(sequence, "sequence", str)

    # PEPTIDE
    if term_char not in sequence and len(sequence.strip()) > 1:
        return sequence.strip()
    if term_char in sequence:
        parts = [part.strip() for part in sequence.split(term_char)]
        # K.PEPTPIDE.P.EP <- wrong format
        if len(parts) > 3:
            raise RuntimeError(f"Could not parse peptide from sequence {sequence}!")
        # K.PEPTIDE.R
        if len(parts) == 3 and len(parts[1]) > 1:
            return parts[1]
        if len(parts) == 2:
            # PEPTIDE.R
            if len(parts[0]) > 1 and len(parts[1]) == 1:
                return parts[0]
            # K.PEPTIDE
            if len(parts[1]) > 1 and len(parts[0]) == 1:
                return parts[1]
    # if none of these cases match, raise error
    raise RuntimeError(f"Could not parse peptide from sequence {sequence}!")
    return "err"


def parse_modifications_from_xi_sequence(sequence: str) -> Dict[int, str]:
    r"""Parses all post-translational-modifications from a peptide sequence as reported by xiFDR.

    Parses all post-translational-modifications from a peptide sequence as reported by xiFDR. This assumes
    that amino acids are given in upper case letters and post-translational-modifications in lower case letters.
    The parsed modifications are returned as a dictionary that maps their position in the sequence (1-based) to
    their xiFDR annotation (``SYMBOLEXT``), for example ``"cm"`` or ``"ox"``.

    Parameters
    ----------
    sequence : str
        The peptide sequence as given by xiFDR.

    Returns
    -------
    dict of int, str
        Dictionary that maps modifications (values) to their respective positions in the peptide sequence (1-based)
        (keys). The modifications are given in xiFDR annotation style (``SYMBOLEXT``) which is the lower letter
        modification code, for example ``"cm"`` for carbamidomethylation.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed.

    Examples
    --------
    >>> from pyXLMS.parser import parse_modifications_from_xi_sequence
    >>> seq1 = "KIECcmFDSVEISGVEDR"
    >>> parse_modifications_from_xi_sequence(seq1)
    {4: 'cm'}

    >>> from pyXLMS.parser import parse_modifications_from_xi_sequence
    >>> seq2 = "KIECcmFDSVEMoxISGVEDR"
    >>> parse_modifications_from_xi_sequence(seq2)
    {4: 'cm', 10: 'ox'}

    >>> from pyXLMS.parser import parse_modifications_from_xi_sequence
    >>> seq3 = "KIECcmFDSVEISGVEDRMox"
    >>> parse_modifications_from_xi_sequence(seq3)
    {4: 'cm', 17: 'ox'}

    >>> from pyXLMS.parser import parse_modifications_from_xi_sequence
    >>> seq4 = "CcmKIECcmFDSVEISGVEDRMox"
    >>> parse_modifications_from_xi_sequence(seq4)
    {1: 'cm', 5: 'cm', 18: 'ox'}
    """
    ## check input
    _ok = check_input(sequence, "sequence", str)

    modifications = dict()
    pos = 0
    current_mod = ""
    for i, aa in enumerate(str(sequence).strip()):
        if aa.isupper():
            pos += 1
            current_mod = ""
        else:
            current_mod += aa
            if (i + 1 >= len(sequence)) or (sequence[i + 1].isupper()):
                if pos in modifications:
                    raise RuntimeError(
                        f"Modification at position {pos} already exists!"
                    )
                modifications[pos] = current_mod
    return modifications


def __parse_int(value: Any) -> int:
    r"""Parses an integer from the given value.

    Parses an integer from the given value. If it is a string it will try to replace any comma
    (thousands seperator) with an empty string.

    Parameters
    ----------
    value : any
        The value to be converted to int.

    Returns
    -------
    int
        The converted integer value.
    """
    if isinstance(value, str):
        return int(value.replace(",", ""))
    return int(value)


def __parse_float(value: Any) -> float:
    r"""Parses a float from the given value.

    Parses a float from the given value. If it is a string it will try to replace any comma
    (thousands seperator) with an empty string.

    Parameters
    ----------
    value : any
        The value to be converted to float.

    Returns
    -------
    float
        The converted float value.
    """
    if isinstance(value, str):
        return float(value.replace(",", ""))
    return float(value)


def __parse_xisearch_modifications(
    row: pd.Series,
    alpha: bool,
    modifications: Dict[str, Tuple[str, float]] = XI_MODIFICATION_MAPPING,
    ignore_errors: bool = False,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[int, Tuple[str, float]]:
    r"""Returns the corresponding modifications object for a crosslink-spectrum-match from xiSearch.

    Parameters
    ----------
    row : pandas.Series
        One row/crosslink-spectrum-match of the xiSearch result file.
    alpha : bool
        Whether to parse modifications from the alpha peptide or - if ``False`` - from the beta peptide.
    modifications : dict of str, tuple, default = ``constants.XI_MODIFICATION_MAPPING``
        Mapping of xi sequence elements (e.g. ``"cm"``) to their modifications (e.g. ``("Carbamidomethyl", 57.021464)``).
    ignore_errors : bool, default = False
        If modifications that are not given in parameter 'modifications' should raise an error or not. By default an error is
        raised if an unknown modification is encountered. If ``True`` modifications that are unknown are encoded with the xi
        shortcode (``SYMBOLEXT``) and ``float("nan")`` modification mass.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict of int, tuple
        The ``pyXLMS`` specific modifications object, a dictionary that maps positions to their corresponding modifications and their
        monoisotopic masses.

    Raises
    ------
    RuntimeError
        If the parsed modifications and positions are not of the same length.
    RuntimeError
        If multiple modifications on the same residue are parsed (only for ``verbose = 2``).
    KeyError
        If an unknown modification is encountered.

    Notes
    -----
    This function should not be called directly, it is called from ``__read_xisearch()``.
    """

    # EXAMPLE VALUES
    # Modifications2            Mox;Mox
    # ModificationPositions2    5;7
    # helper function that changes ``SYMBOL`` to ``SYMBOLEXT``
    def preprocess_mod(mod: str) -> str:
        return "".join([c for c in mod if not c.isupper()]).strip()

    crosslinker = str(row["Crosslinker"]).strip()
    crosslinker_mass = __parse_float(row["CrosslinkerMass"])
    parsed_modifications = dict()
    # parse from Modifications
    if alpha:
        parsed_modifications[__parse_int(row["Link1"])] = (
            crosslinker,
            crosslinker_mass,
        )
        if not pd.isna(row["Modifications1"]):  # pyright: ignore [reportGeneralTypeIssues]
            if ";" in str(row["Modifications1"]):
                mods = [
                    preprocess_mod(mod) for mod in str(row["Modifications1"]).split(";")
                ]
                positions = [
                    __parse_int(pos)
                    for pos in str(row["ModificationPositions1"]).split(";")
                ]
                if len(mods) != len(positions):
                    err_str = "Parsed modifications and their positions are not of the same length!\n"
                    err_str += f"Parsed modifications: {row['Modifications1']}; Parsed positions: {row['ModificationPositions1']}\n"
                    err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                    raise RuntimeError(err_str)
                for i in range(len(mods)):
                    if positions[i] in parsed_modifications:
                        err_str = (
                            f"Modification at position {positions[i]} already exists!\n"
                        )
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        if verbose == 1:
                            warnings.warn(RuntimeWarning(err_str))
                        elif verbose == 2:
                            raise RuntimeError(err_str)
                        try:
                            t1 = parsed_modifications[positions[i]][0] + (
                                "," + modifications[mods[i]][0]
                            )
                            t2 = (
                                parsed_modifications[positions[i]][1]
                                + modifications[mods[i]][1]
                            )
                            parsed_modifications[positions[i]] = (t1, t2)
                        except KeyError:
                            if ignore_errors:
                                t1 = (
                                    parsed_modifications[positions[i]][0]
                                    + ","
                                    + mods[i]
                                )
                                t2 = parsed_modifications[positions[i]][1] + float(
                                    "nan"
                                )
                                parsed_modifications[positions[i]] = (t1, t2)
                            else:
                                err_str = f"Key {mods[i]} not found in parameter 'modifications'. Are you missing a modification?\n"
                                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                                raise KeyError(err_str)
                    else:
                        try:
                            parsed_modifications[positions[i]] = (
                                modifications[mods[i]][0],
                                modifications[mods[i]][1],
                            )
                        except KeyError:
                            if ignore_errors:
                                parsed_modifications[positions[i]] = (
                                    mods[i],
                                    float("nan"),
                                )
                            else:
                                err_str = f"Key {mods[i]} not found in parameter 'modifications'. Are you missing a modification?\n"
                                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                                raise KeyError(err_str)
            else:
                mod = preprocess_mod(str(row["Modifications1"]))
                pos = __parse_int(row["ModificationPositions1"])
                if pos in parsed_modifications:
                    err_str = f"Modification at position {pos} already exists!\n"
                    err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                    if verbose == 1:
                        warnings.warn(RuntimeWarning(err_str))
                    elif verbose == 2:
                        raise RuntimeError(err_str)
                    try:
                        t1 = parsed_modifications[pos][0] + "," + modifications[mod][0]
                        t2 = parsed_modifications[pos][1] + modifications[mod][1]
                        parsed_modifications[pos] = (t1, t2)
                    except KeyError:
                        if ignore_errors:
                            t1 = parsed_modifications[pos][0] + "," + mod
                            t2 = parsed_modifications[pos][1] + float("nan")
                            parsed_modifications[pos] = (t1, t2)
                        else:
                            err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                            err_str += (
                                f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                            )
                            raise KeyError(err_str)
                else:
                    try:
                        parsed_modifications[pos] = (
                            modifications[mod][0],
                            modifications[mod][1],
                        )
                    except KeyError:
                        if ignore_errors:
                            parsed_modifications[pos] = (
                                mod,
                                float("nan"),
                            )
                        else:
                            err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                            err_str += (
                                f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                            )
                            raise KeyError(err_str)
    else:
        parsed_modifications[__parse_int(row["Link2"])] = (
            crosslinker,
            crosslinker_mass,
        )
        if not pd.isna(row["Modifications2"]):  # pyright: ignore [reportGeneralTypeIssues]
            if ";" in str(row["Modifications2"]):
                mods = [
                    preprocess_mod(mod) for mod in str(row["Modifications2"]).split(";")
                ]
                positions = [
                    __parse_int(pos)
                    for pos in str(row["ModificationPositions2"]).split(";")
                ]
                if len(mods) != len(positions):
                    err_str = "Parsed modifications and their positions are not of the same length!\n"
                    err_str += f"Parsed modifications: {row['Modifications2']}; Parsed positions: {row['ModificationPositions2']}\n"
                    err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                    raise RuntimeError(err_str)
                for i in range(len(mods)):
                    if positions[i] in parsed_modifications:
                        err_str = (
                            f"Modification at position {positions[i]} already exists!\n"
                        )
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        if verbose == 1:
                            warnings.warn(RuntimeWarning(err_str))
                        elif verbose == 2:
                            raise RuntimeError(err_str)
                        try:
                            t1 = parsed_modifications[positions[i]][0] + (
                                "," + modifications[mods[i]][0]
                            )
                            t2 = (
                                parsed_modifications[positions[i]][1]
                                + modifications[mods[i]][1]
                            )
                            parsed_modifications[positions[i]] = (t1, t2)
                        except KeyError:
                            if ignore_errors:
                                t1 = parsed_modifications[positions[i]][0] + (
                                    "," + mods[i]
                                )
                                t2 = parsed_modifications[positions[i]][1] + float(
                                    "nan"
                                )
                                parsed_modifications[positions[i]] = (t1, t2)
                            else:
                                err_str = f"Key {mods[i]} not found in parameter 'modifications'. Are you missing a modification?\n"
                                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                                raise KeyError(err_str)
                    else:
                        try:
                            parsed_modifications[positions[i]] = (
                                modifications[mods[i]][0],
                                modifications[mods[i]][1],
                            )
                        except KeyError:
                            if ignore_errors:
                                parsed_modifications[positions[i]] = (
                                    mods[i],
                                    float("nan"),
                                )
                            else:
                                err_str = f"Key {mods[i]} not found in parameter 'modifications'. Are you missing a modification?\n"
                                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                                raise KeyError(err_str)
            else:
                mod = preprocess_mod(str(row["Modifications2"]))
                pos = __parse_int(row["ModificationPositions2"])
                if pos in parsed_modifications:
                    err_str = f"Modification at position {pos} already exists!\n"
                    err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                    if verbose == 1:
                        warnings.warn(RuntimeWarning(err_str))
                    elif verbose == 2:
                        raise RuntimeError(err_str)
                    try:
                        t1 = parsed_modifications[pos][0] + "," + modifications[mod][0]
                        t2 = parsed_modifications[pos][1] + modifications[mod][1]
                        parsed_modifications[pos] = (t1, t2)
                    except KeyError:
                        if ignore_errors:
                            t1 = parsed_modifications[pos][0] + "," + mod
                            t2 = parsed_modifications[pos][1] + float("nan")
                            parsed_modifications[pos] = (t1, t2)
                        else:
                            err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                            err_str += (
                                f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                            )
                            raise KeyError(err_str)
                else:
                    try:
                        parsed_modifications[pos] = (
                            modifications[mod][0],
                            modifications[mod][1],
                        )
                    except KeyError:
                        if ignore_errors:
                            parsed_modifications[pos] = (
                                mod,
                                float("nan"),
                            )
                        else:
                            err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                            err_str += (
                                f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                            )
                            raise KeyError(err_str)
    # parse from sequence (because fixed modifcations are not reported in Modifications)
    if alpha:
        modified_sequence = parse_peptide(str(row["Peptide1"]).strip())
        mods_from_sequence = parse_modifications_from_xi_sequence(modified_sequence)
        for pos, mod in mods_from_sequence.items():
            if pos in parsed_modifications:
                err_str = f"Modification at position {pos} already exists!\n"
                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                if verbose == 1:
                    warnings.warn(RuntimeWarning(err_str))
                elif verbose == 2:
                    raise RuntimeError(err_str)
                mod_mapped = None
                try:
                    mod_mapped = modifications[mod]
                except KeyError:
                    if ignore_errors:
                        mod_mapped = (mod, float("nan"))
                    else:
                        err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        raise KeyError(err_str)
                if mod_mapped is not None and isinstance(mod_mapped, tuple):
                    if mod_mapped[0] not in parsed_modifications[pos][0]:
                        parsed_modifications[pos] = (
                            parsed_modifications[pos][0] + "," + mod_mapped[0],
                            parsed_modifications[pos][1] + mod_mapped[1],
                        )
            else:
                mod_mapped = None
                try:
                    mod_mapped = modifications[mod]
                except KeyError:
                    if ignore_errors:
                        mod_mapped = (mod, float("nan"))
                    else:
                        err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        raise KeyError(err_str)
                if mod_mapped is not None and isinstance(mod_mapped, tuple):
                    parsed_modifications[pos] = mod_mapped
    else:
        modified_sequence = parse_peptide(str(row["Peptide2"]).strip())
        mods_from_sequence = parse_modifications_from_xi_sequence(modified_sequence)
        for pos, mod in mods_from_sequence.items():
            if pos in parsed_modifications:
                err_str = f"Modification at position {pos} already exists!\n"
                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                if verbose == 1:
                    warnings.warn(RuntimeWarning(err_str))
                elif verbose == 2:
                    raise RuntimeError(err_str)
                mod_mapped = None
                try:
                    mod_mapped = modifications[mod]
                except KeyError:
                    if ignore_errors:
                        mod_mapped = (mod, float("nan"))
                    else:
                        err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        raise KeyError(err_str)
                if mod_mapped is not None and isinstance(mod_mapped, tuple):
                    if mod_mapped[0] not in parsed_modifications[pos][0]:
                        parsed_modifications[pos] = (
                            parsed_modifications[pos][0] + "," + mod_mapped[0],
                            parsed_modifications[pos][1] + mod_mapped[1],
                        )
            else:
                mod_mapped = None
                try:
                    mod_mapped = modifications[mod]
                except KeyError:
                    if ignore_errors:
                        mod_mapped = (mod, float("nan"))
                    else:
                        err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        raise KeyError(err_str)
                if mod_mapped is not None and isinstance(mod_mapped, tuple):
                    parsed_modifications[pos] = mod_mapped
    return parsed_modifications


def __read_xisearch(
    data: pd.DataFrame,
    decoy_prefix: str,
    parse_modifications: bool,
    modifications: Dict[str, Tuple[str, float]],
    ignore_errors: bool,
    verbose: Literal[0, 1, 2],
) -> List[Dict[str, Any]]:
    r"""Reads a xiSearch pandas dataframe and returns a list of crosslink-spectrum-matches.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe of a xiSearch result ``.csv`` file read with pandas.
    decoy_prefix : str
        The prefix that indicates that a protein is from the decoy database.
    parse_modifications : bool
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications : dict of str, tuple
        Mapping of xi sequence elements (e.g. ``"cm"``) to their modifications (e.g. ``("Carbamidomethyl", 57.021464)``).
    ignore_errors : bool
        If modifications that are not given in parameter 'modifications' should raise an error or not. By default an error is
        raised if an unknown modification is encountered. If ``True`` modifications that are unknown are encoded with the xi
        shortcode (``SYMBOLEXT``) and ``float("nan")`` modification mass.
    verbose : 0, 1, or 2
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    list of dict
        The read crosslink-spectrum-matches.

    Notes
    -----
    This function should not be called directly, it is called from ``read_xi()``.
    """
    # remove monolinks
    xl = data.dropna(axis=0, subset=["BasePeptide2"])
    # create csms list
    csms = list()
    # create csms
    for i, row in tqdm(
        xl.iterrows(), total=xl.shape[0], desc="Reading xiSearch CSMs..."
    ):
        csm = create_csm(
            peptide_a=format_sequence(str(row["BasePeptide1"])),
            modifications_a=__parse_xisearch_modifications(
                row, True, modifications, ignore_errors, verbose
            )
            if parse_modifications
            else None,
            xl_position_peptide_a=__parse_int(row["Link1"]),
            proteins_a=[
                p.strip()
                if p.strip()[: len(decoy_prefix)] != decoy_prefix
                else p.strip()[len(decoy_prefix) :]
                for p in str(row["Protein1"]).split(";")
            ],
            xl_position_proteins_a=[
                __parse_int(__parse_float(p))
                for p in str(row["ProteinLink1"]).split(";")
            ],
            pep_position_proteins_a=[
                __parse_int(__parse_float(p)) for p in str(row["Start1"]).split(";")
            ],
            score_a=__parse_float(row["Pep1Score"]),
            decoy_a=get_bool_from_value(int(row["Protein1decoy"])),
            peptide_b=format_sequence(str(row["BasePeptide2"])),
            modifications_b=__parse_xisearch_modifications(
                row, False, modifications, ignore_errors, verbose
            )
            if parse_modifications
            else None,
            xl_position_peptide_b=__parse_int(row["Link2"]),
            proteins_b=[
                p.strip()
                if p.strip()[: len(decoy_prefix)] != decoy_prefix
                else p.strip()[len(decoy_prefix) :]
                for p in str(row["Protein2"]).split(";")
            ],
            xl_position_proteins_b=[
                __parse_int(__parse_float(p))
                for p in str(row["ProteinLink2"]).split(";")
            ],
            pep_position_proteins_b=[
                __parse_int(__parse_float(p)) for p in str(row["Start2"]).split(";")
            ],
            score_b=__parse_float(row["Pep2Score"]),
            decoy_b=get_bool_from_value(int(row["Protein2decoy"])),
            score=__parse_float(row["match score"]),
            spectrum_file=str(row["peakListFileName"]).strip(),
            scan_nr=__parse_int(row["Scan"]),
            charge=__parse_int(row["PrecoursorCharge"]),
            rt=None,
            im_cv=None,
            additional_information={
                "source": __serialize_pandas_series(row),
                "spectrum quality score": __parse_float(row["spectrum quality score"]),
            },
        )
        csms.append(csm)
    return csms


def __parse_xifdr_modifications(
    row: pd.Series,
    alpha: bool,
    modifications: Dict[str, Tuple[str, float]] = XI_MODIFICATION_MAPPING,
    ignore_errors: bool = False,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[int, Tuple[str, float]]:
    r"""Returns the corresponding modifications object for a crosslink-spectrum-match from xiFDR.

    Parameters
    ----------
    row : pandas.Series
        One row/crosslink-spectrum-match of the xiFDR CSM result file.
    alpha : bool
        Whether to parse modifications from the alpha peptide or - if ``False`` - from the beta peptide.
    modifications : dict of str, tuple, default = ``constants.XI_MODIFICATION_MAPPING``
        Mapping of xi sequence elements (e.g. ``"cm"``) to their modifications (e.g. ``("Carbamidomethyl", 57.021464)``).
    ignore_errors : bool, default = False
        If modifications that are not given in parameter 'modifications' should raise an error or not. By default an error is
        raised if an unknown modification is encountered. If ``True`` modifications that are unknown are encoded with the xi
        shortcode (``SYMBOLEXT``) and ``float("nan")`` modification mass.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict of int, tuple
        The ``pyXLMS`` specific modifications object, a dictionary that maps positions to their corresponding modifications and their
        monoisotopic masses.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed (only if ``verbose = 2``).
    KeyError
        If an unknown modification is encountered.

    Notes
    -----
    This function should not be called directly, it is called from ``__read_xifdr_csms()``.
    """
    crosslinker = str(row["Crosslinker"]).strip()
    crosslinker_mass = __parse_float(row["CrosslinkerModMass"])
    parsed_modifications = dict()
    if alpha:
        parsed_modifications[__parse_int(row["LinkPos1"])] = (
            crosslinker,
            crosslinker_mass,
        )
        for pos, mod in parse_modifications_from_xi_sequence(
            str(row["PepSeq1"]).strip()
        ).items():
            if pos in parsed_modifications:
                err_str = f"Modification at position {pos} already exists!\n"
                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                if verbose == 1:
                    warnings.warn(RuntimeWarning(err_str))
                elif verbose == 2:
                    raise RuntimeError(err_str)
                try:
                    t1 = parsed_modifications[pos][0] + "," + modifications[mod][0]
                    t2 = parsed_modifications[pos][1] + modifications[mod][1]
                    parsed_modifications[pos] = (t1, t2)
                except KeyError:
                    if ignore_errors:
                        t1 = parsed_modifications[pos][0] + "," + mod
                        t2 = parsed_modifications[pos][1] + float("nan")
                        parsed_modifications[pos] = (t1, t2)
                    else:
                        err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        raise KeyError(err_str)
            try:
                parsed_modifications[pos] = (
                    modifications[mod][0],
                    modifications[mod][1],
                )
            except KeyError:
                if ignore_errors:
                    parsed_modifications[pos] = (mod, float("nan"))
                else:
                    err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                    err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                    raise KeyError(err_str)
    else:
        parsed_modifications[__parse_int(row["LinkPos2"])] = (
            crosslinker,
            crosslinker_mass,
        )
        for pos, mod in parse_modifications_from_xi_sequence(
            str(row["PepSeq2"]).strip()
        ).items():
            if pos in parsed_modifications:
                err_str = f"Modification at position {pos} already exists!\n"
                err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                if verbose == 1:
                    warnings.warn(RuntimeWarning(err_str))
                elif verbose == 2:
                    raise RuntimeError(err_str)
                try:
                    t1 = parsed_modifications[pos][0] + "," + modifications[mod][0]
                    t2 = parsed_modifications[pos][1] + modifications[mod][1]
                    parsed_modifications[pos] = (t1, t2)
                except KeyError:
                    if ignore_errors:
                        t1 = parsed_modifications[pos][0] + "," + mod
                        t2 = parsed_modifications[pos][1] + float("nan")
                        parsed_modifications[pos] = (t1, t2)
                    else:
                        err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                        err_str += (
                            f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                        )
                        raise KeyError(err_str)
            try:
                parsed_modifications[pos] = (
                    modifications[mod][0],
                    modifications[mod][1],
                )
            except KeyError:
                if ignore_errors:
                    parsed_modifications[pos] = (mod, float("nan"))
                else:
                    err_str = f"Key {mod} not found in parameter 'modifications'. Are you missing a modification?\n"
                    err_str += f"CSM ScanId: {row['ScanId']}; CSM Scan: {row['Scan']}"
                    raise KeyError(err_str)
    return parsed_modifications


def __read_xifdr_csms(
    data: pd.DataFrame,
    decoy_prefix: str,
    parse_modifications: bool,
    modifications: Dict[str, Tuple[str, float]],
    ignore_errors: bool,
    verbose: Literal[0, 1, 2],
) -> List[Dict[str, Any]]:
    r"""Reads a xiFDR CSM pandas dataframe and returns a list of crosslink-spectrum-matches.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe of a xiFDR CSM result ``.csv`` file read with pandas.
    decoy_prefix : str
        The prefix that indicates that a protein is from the decoy database.
    parse_modifications : bool
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications : dict of str, tuple
        Mapping of xi sequence elements (e.g. ``"cm"``) to their modifications (e.g. ``("Carbamidomethyl", 57.021464)``).
    ignore_errors : bool
        If modifications that are not given in parameter 'modifications' should raise an error or not. By default an error is
        raised if an unknown modification is encountered. If ``True`` modifications that are unknown are encoded with the xi
        shortcode (``SYMBOLEXT``) and ``float("nan")`` modification mass.
    verbose : 0, 1, or 2
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    list of dict
        The read crosslink-spectrum-matches.

    Notes
    -----
    This function should not be called directly, it is called from ``read_xi()``.
    """
    # create csms list
    csms = list()
    # create csms
    for i, row in tqdm(
        data.iterrows(), total=data.shape[0], desc="Reading xiFDR CSMs..."
    ):
        csm = create_csm(
            peptide_a=format_sequence(str(row["PepSeq1"])),
            modifications_a=__parse_xifdr_modifications(
                row, True, modifications, ignore_errors, verbose
            )
            if parse_modifications
            else None,
            xl_position_peptide_a=__parse_int(row["LinkPos1"]),
            proteins_a=[
                p.strip()
                if p.strip()[: len(decoy_prefix)] != decoy_prefix
                else p.strip()[len(decoy_prefix) :]
                for p in str(row["Protein1"]).split(";")
            ],
            xl_position_proteins_a=[
                __parse_int(p) for p in str(row["ProteinLinkPos1"]).split(";")
            ],
            pep_position_proteins_a=[
                __parse_int(p) for p in str(row["PepPos1"]).split(";")
            ],
            score_a=None,
            decoy_a=get_bool_from_value(row["Decoy1"]),
            peptide_b=format_sequence(str(row["PepSeq2"])),
            modifications_b=__parse_xifdr_modifications(
                row, False, modifications, ignore_errors, verbose
            )
            if parse_modifications
            else None,
            xl_position_peptide_b=__parse_int(row["LinkPos2"]),
            proteins_b=[
                p.strip()
                if p.strip()[: len(decoy_prefix)] != decoy_prefix
                else p.strip()[len(decoy_prefix) :]
                for p in str(row["Protein2"]).split(";")
            ],
            xl_position_proteins_b=[
                __parse_int(p) for p in str(row["ProteinLinkPos2"]).split(";")
            ],
            pep_position_proteins_b=[
                __parse_int(p) for p in str(row["PepPos2"]).split(";")
            ],
            score_b=None,
            decoy_b=get_bool_from_value(row["Decoy2"]),
            score=__parse_float(row["Score"]),
            spectrum_file=str(row["PeakListFileName"]).strip(),
            scan_nr=__parse_int(row["scan"]),
            charge=__parse_int(row["exp charge"]),
            rt=None,
            im_cv=None,
            additional_information={"source": __serialize_pandas_series(row)},
        )
        csms.append(csm)
    return csms


def __read_xifdr_crosslinks(
    data: pd.DataFrame, decoy_prefix: str
) -> List[Dict[str, Any]]:
    r"""Reads a xiFDR Links pandas dataframe and returns a list of crosslinks.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe of a xiFDR Links result ``.csv`` file read with pandas.
    decoy_prefix : str
        The prefix that indicates that a protein is from the decoy database.

    Returns
    -------
    list of dict
        The read crosslinks.

    Notes
    -----
    This function should not be called directly, it is called from ``read_xi()``.
    """
    # create crosslink list
    crosslinks = list()
    # create crosslinks
    for i, row in tqdm(
        data.iterrows(), total=data.shape[0], desc="Reading xiFDR crosslinks..."
    ):
        psmid = str(row["PSMIDs"]).split(";")[0]
        s1 = psmid.split("P1_")[1].split(" ")[0]
        p1 = parse_peptide(s1)
        s2 = psmid.split("P2_")[1].split(" ")[0]
        p2 = parse_peptide(s2)
        pos1 = __parse_int(psmid.split("P2_")[1].split(" ")[1])
        pos2 = __parse_int(psmid.split("P2_")[1].split(" ")[2])
        crosslink = create_crosslink(
            peptide_a=format_sequence(p1),
            xl_position_peptide_a=pos1,
            proteins_a=[
                p.strip()
                if p.strip()[: len(decoy_prefix)] != decoy_prefix
                else p.strip()[len(decoy_prefix) :]
                for p in str(row["Protein1"]).split(";")
            ],
            xl_position_proteins_a=[
                __parse_int(p) for p in str(row["fromSite"]).split(";")
            ],
            decoy_a=get_bool_from_value(row["Decoy1"]),
            peptide_b=format_sequence(p2),
            xl_position_peptide_b=pos2,
            proteins_b=[
                p.strip()
                if p.strip()[: len(decoy_prefix)] != decoy_prefix
                else p.strip()[len(decoy_prefix) :]
                for p in str(row["Protein2"]).split(";")
            ],
            xl_position_proteins_b=[
                __parse_int(p) for p in str(row["ToSite"]).split(";")
            ],
            decoy_b=get_bool_from_value(row["Decoy2"]),
            score=__parse_float(row["Score"]),
            additional_information={"source": __serialize_pandas_series(row)},
        )
        crosslinks.append(crosslink)
    return crosslinks


def read_xi(
    files: str | List[str] | BinaryIO,
    decoy_prefix: Optional[str] = "auto",
    parse_modifications: bool = True,
    modifications: Dict[str, Tuple[str, float]] = XI_MODIFICATION_MAPPING,
    sep: str = ",",
    decimal: str = ".",
    ignore_errors: bool = False,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[str, Any]:
    r"""Read a xiSearch/xiFDR result file.

    Reads a xiSearch crosslink-spectrum-matches result file or a xiFDR crosslink-spectrum-matches
    result file or crosslink result file in ``.csv`` format and returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the xiSearch/xiFDR result file(s) or a file-like object/stream.
    decoy_prefix : str, or None, default = "auto"
        The prefix that indicates that a protein is from the decoy database.
        If "auto" or None it will use the default for each xi file type.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications : dict of str, tuple, default = ``constants.XI_MODIFICATION_MAPPING``
        Mapping of xi sequence elements (e.g. ``"cm"``) to their modifications (e.g. ``("Carbamidomethyl", 57.021464)``).
        This corresponds to the ``SYMBOLEXT`` field, or the ``SYMBOL`` field minus the amino acid in the xiSearch config.
    sep : str, default = ","
        Seperator used in the ``.csv`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.
    ignore_errors : bool, default = False
        If modifications that are not given in parameter 'modifications' should raise an error or not. By default an error is
        raised if an unknown modification is encountered. If ``True`` modifications that are unknown are encoded with the xi
        shortcode (``SYMBOLEXT``) and ``float("nan")`` modification mass.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) contain no crosslinks or crosslink-spectrum-matches.
    TypeError
        If parameter verbose was not set correctly.

    Notes
    -----
    Uses ``Pep1Score`` as the score for the alpha peptide, ``Pep2Score`` as the score of the
    beta peptide, and ``match score`` as the score of the crosslink-spectrum-match for xiSearch
    crosslink-spectrum-matches. Uses ``Score`` as the score of the crosslink-spectrum-match for
    xiFDR crosslink-spectrum-matches, alpha and beta peptide scores are ``None`` for xiFDR
    crosslink-spectrum-matches. Uses ``Score`` as the score of the crosslink for xiFDR crosslinks.
    For reference, see here:
    `github.com/Rappsilber-Laboratory/XiSearch <https://github.com/Rappsilber-Laboratory/XiSearch/discussions/126>`_.

    Examples
    --------
    >>> from pyXLMS.parser import read_xi
    >>> csms_from_xiSearch = read_xi("data/xi/r1_Xi1.7.6.7.csv")

    >>> from pyXLMS.parser import read_xi
    >>> csms_from_xiFDR = read_xi("data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv")

    >>> from pyXLMS.parser import read_xi
    >>> crosslinks_from_xiFDR = read_xi("data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv")
    """
    ## check input
    _ok = (
        check_input(decoy_prefix, "decoy_prefix", str)
        if decoy_prefix is not None
        else True
    )
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = check_input(modifications, "modifications", dict, tuple)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    _ok = check_input(ignore_errors, "ignore_errors", bool)
    _ok = check_input(verbose, "verbose", int)
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")

    ## data structures
    crosslinks = list()
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    for input in inputs:
        ## reading data
        data = pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
        ## detect input file type
        xi_file_type = detect_xi_filetype(data)
        ## set decoy prefix
        if decoy_prefix is None or decoy_prefix == "auto":
            decoy_prefix = "REV_" if xi_file_type == "xisearch" else "decoy:"
        ## process data
        if xi_file_type == "xifdr_csms":
            csms += __read_xifdr_csms(
                data,
                decoy_prefix,
                parse_modifications,
                modifications,
                ignore_errors,
                verbose,
            )
        elif xi_file_type == "xifdr_crosslinks":
            crosslinks += __read_xifdr_crosslinks(data, decoy_prefix)
        else:
            csms += __read_xisearch(
                data,
                decoy_prefix,
                parse_modifications,
                modifications,
                ignore_errors,
                verbose,
            )

    ## check results
    if len(crosslinks) + len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches or crosslinks were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="xiSearch/xiFDR",
        csms=csms if len(csms) > 0 else None,
        crosslinks=crosslinks if len(crosslinks) > 0 else None,
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


PD = "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3.pdResult"
DECOY = False


def test1():
    from pyXLMS import parser as p
    from pyXLMS.transform import modifications_to_str as mts

    parser_result = p.read_xlinkx(PD)
    assert parser_result["data_type"] == "parser_result"
    assert parser_result["completeness"] == "full"
    assert parser_result["search_engine"] == "XlinkX"
    assert parser_result["crosslink-spectrum-matches"] is not None
    assert parser_result["crosslinks"] is not None

    csms = parser_result["crosslink-spectrum-matches"]
    assert len(csms) == 841 + 9 if DECOY else 841

    crosslinks = parser_result["crosslinks"]
    assert len(crosslinks) == 262 if DECOY else 253

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "GQKNSR"
    assert mts(csm["alpha_modifications"]) == "(3:[DSSO|158.00376])"
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [779]
    assert csm["alpha_proteins_peptide_positions"] == [777]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "GYKEVK"
    assert mts(csm["beta_modifications"]) == "(3:[DSSO|158.00376])"
    assert csm["beta_peptide_crosslink_position"] == 3
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [1192]
    assert csm["beta_proteins_peptide_positions"] == [1190]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(125.007057923049)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_Lumos_DSSO_MS3.raw"
    assert csm["scan_nr"] == 4313
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(16.0316734313965 * 60.0)
    assert csm["ion_mobility"] is None

    if DECOY:
        csm = csms[-1]
        assert csm["data_type"] == "crosslink-spectrum-match"
        assert csm["completeness"] == "partial"
        assert csm["alpha_peptide"] == "LNDWTLDSASKK"
        assert mts(csm["alpha_modifications"]) == "(11:[DSSO|158.00376])"
        assert csm["alpha_peptide_crosslink_position"] == 11
        assert csm["alpha_proteins"] == ["decoy_98"]
        assert csm["alpha_proteins_crosslink_positions"] == [235]
        assert csm["alpha_proteins_peptide_positions"] == [225]
        assert csm["alpha_score"] is None
        assert csm["alpha_decoy"]
        assert csm["beta_peptide"] == "NKDSR"
        assert mts(csm["beta_modifications"]) == "(2:[DSSO|158.00376])"
        assert csm["beta_peptide_crosslink_position"] == 2
        assert csm["beta_proteins"] == ["decoy_2"]
        assert csm["beta_proteins_crosslink_positions"] == [505]
        assert csm["beta_proteins_peptide_positions"] == [504]
        assert csm["beta_score"] is None
        assert csm["beta_decoy"]
        assert csm["crosslink_type"] == "inter"
        assert csm["score"] == pytest.approx(4.15420412699015)
        assert csm["spectrum_file"] == "XLpeplib_Beveridge_Lumos_DSSO_MS3.raw"
        assert csm["scan_nr"] == 43290
        assert csm["charge"] == 3
        assert csm["retention_time"] == pytest.approx(88.1825561523437 * 60.0)
        assert csm["ion_mobility"] is None

    xl = crosslinks[1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "EKIEK"
    assert xl["alpha_peptide_crosslink_position"] == 2
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [443]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "KVTVK"
    assert xl["beta_peptide_crosslink_position"] == 1
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [562]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(119.059334966149)

    if DECOY:
        xl = crosslinks[-1]
        assert xl["data_type"] == "crosslink"
        assert xl["completeness"] == "full"
        assert xl["alpha_peptide"] == "LNDWTLDSASKK"
        assert xl["alpha_peptide_crosslink_position"] == 11
        assert xl["alpha_proteins"] == ["decoy_98"]
        assert xl["alpha_proteins_crosslink_positions"] == [235]
        assert xl["alpha_decoy"]
        assert xl["beta_peptide"] == "NKDSR"
        assert xl["beta_peptide_crosslink_position"] == 2
        assert xl["beta_proteins"] == ["decoy_2"]
        assert xl["beta_proteins_crosslink_positions"] == [505]
        assert xl["beta_decoy"]
        assert xl["crosslink_type"] == "inter"
        assert xl["score"] == pytest.approx(4.15420412699015)


def test2():
    from pyXLMS import parser as p
    from pyXLMS.transform import modifications_to_str as mts

    parser_result = p.read_xlinkx([PD, PD])
    assert parser_result["data_type"] == "parser_result"
    assert parser_result["completeness"] == "full"
    assert parser_result["search_engine"] == "XlinkX"
    assert parser_result["crosslink-spectrum-matches"] is not None
    assert parser_result["crosslinks"] is not None

    csms = parser_result["crosslink-spectrum-matches"]
    assert len(csms) == (841 + 9) * 2 if DECOY else 841 * 2

    crosslinks = parser_result["crosslinks"]
    assert len(crosslinks) == 262 * 2 if DECOY else 253 * 2

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "GQKNSR"
    assert mts(csm["alpha_modifications"]) == "(3:[DSSO|158.00376])"
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [779]
    assert csm["alpha_proteins_peptide_positions"] == [777]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "GYKEVK"
    assert mts(csm["beta_modifications"]) == "(3:[DSSO|158.00376])"
    assert csm["beta_peptide_crosslink_position"] == 3
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [1192]
    assert csm["beta_proteins_peptide_positions"] == [1190]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(125.007057923049)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_Lumos_DSSO_MS3.raw"
    assert csm["scan_nr"] == 4313
    assert csm["charge"] == 3
    assert csm["retention_time"] == pytest.approx(16.0316734313965 * 60.0)
    assert csm["ion_mobility"] is None

    if DECOY:
        csm = csms[-1]
        assert csm["data_type"] == "crosslink-spectrum-match"
        assert csm["completeness"] == "partial"
        assert csm["alpha_peptide"] == "LNDWTLDSASKK"
        assert mts(csm["alpha_modifications"]) == "(11:[DSSO|158.00376])"
        assert csm["alpha_peptide_crosslink_position"] == 11
        assert csm["alpha_proteins"] == ["decoy_98"]
        assert csm["alpha_proteins_crosslink_positions"] == [235]
        assert csm["alpha_proteins_peptide_positions"] == [225]
        assert csm["alpha_score"] is None
        assert csm["alpha_decoy"]
        assert csm["beta_peptide"] == "NKDSR"
        assert mts(csm["beta_modifications"]) == "(2:[DSSO|158.00376])"
        assert csm["beta_peptide_crosslink_position"] == 2
        assert csm["beta_proteins"] == ["decoy_2"]
        assert csm["beta_proteins_crosslink_positions"] == [505]
        assert csm["beta_proteins_peptide_positions"] == [504]
        assert csm["beta_score"] is None
        assert csm["beta_decoy"]
        assert csm["crosslink_type"] == "inter"
        assert csm["score"] == pytest.approx(4.15420412699015)
        assert csm["spectrum_file"] == "XLpeplib_Beveridge_Lumos_DSSO_MS3.raw"
        assert csm["scan_nr"] == 43290
        assert csm["charge"] == 3
        assert csm["retention_time"] == pytest.approx(88.1825561523437 * 60.0)
        assert csm["ion_mobility"] is None

    xl = crosslinks[1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "EKIEK"
    assert xl["alpha_peptide_crosslink_position"] == 2
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [443]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "KVTVK"
    assert xl["beta_peptide_crosslink_position"] == 1
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [562]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(119.059334966149)

    if DECOY:
        xl = crosslinks[-1]
        assert xl["data_type"] == "crosslink"
        assert xl["completeness"] == "full"
        assert xl["alpha_peptide"] == "LNDWTLDSASKK"
        assert xl["alpha_peptide_crosslink_position"] == 11
        assert xl["alpha_proteins"] == ["decoy_98"]
        assert xl["alpha_proteins_crosslink_positions"] == [235]
        assert xl["alpha_decoy"]
        assert xl["beta_peptide"] == "NKDSR"
        assert xl["beta_peptide_crosslink_position"] == 2
        assert xl["beta_proteins"] == ["decoy_2"]
        assert xl["beta_proteins_crosslink_positions"] == [505]
        assert xl["beta_decoy"]
        assert xl["crosslink_type"] == "inter"
        assert xl["score"] == pytest.approx(4.15420412699015)


def test3():
    from pyXLMS import parser as p

    parser_result = p.read_xlinkx("data/xlinkx/Cas9_HeLa_Cyt_r1.pdResult")
    assert parser_result["data_type"] == "parser_result"
    assert parser_result["completeness"] == "full"
    assert parser_result["search_engine"] == "XlinkX"
    assert parser_result["crosslink-spectrum-matches"] is not None
    assert parser_result["crosslinks"] is not None


def test4():
    from pyXLMS import parser as p

    parser_result = p.read_xlinkx("data/xlinkx/Cas9_HeLa_Cyt_r2.pdResult")
    assert parser_result["data_type"] == "parser_result"
    assert parser_result["completeness"] == "full"
    assert parser_result["search_engine"] == "XlinkX"
    assert parser_result["crosslink-spectrum-matches"] is not None
    assert parser_result["crosslinks"] is not None


def test5():
    from pyXLMS import parser as p

    parser_result = p.read_xlinkx("data/xlinkx/Cas9_HeLa_Cyt_r3.pdResult")
    assert parser_result["data_type"] == "parser_result"
    assert parser_result["completeness"] == "full"
    assert parser_result["search_engine"] == "XlinkX"
    assert parser_result["crosslink-spectrum-matches"] is not None
    assert parser_result["crosslinks"] is not None



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

CAS9 = """
GAASMDKKYSIGLAIGTNSVGWAVITDEYKVPSKKFKVLGNTDRHSIKKNLIG
ALLFDSGETAEATRLKRTARRRYTRRKNRICYLQEIFSNEMAKVDDSFFHRLE
ESFLVEEDKKHERHPIFGNIVDEVAYHEKYPTIYHLRKKLVDSTDKADLRLIYL
ALAHMIKFRGHFLIEGDLNPDNSDVDKLFIQLVQTYNQLFEENPINASGVDA
KAILSARLSKSRRLENLIAQLPGEKKNGLFGNLIALSLGLTPNFKSNFDLAE
DAKLQLSKDTYDDDLDNLLAQIGDQYADLFLAAKNLSDAILLSDILRVNTEITKA
PLSASMIKRYDEHHQDLTLLKALVRQQLPEKYKEIFFDQSKNGYAGYIDGG
ASQEEFYKFIKPILEKMDGTEELLVKLNREDLLRKQRTFDNGSIPHQIHLGE
LHAILRRQEDFYPFLKDNREKIEKILTFRIPYYVGPLARGNSRFAWMTRKSEETI
TPWNFEEVVDKGASAQSFIERMTNFDKNLPNEKVLPKHSLLYEYFTVYNELT
KVKYVTEGMRKPAFLSGEQKKAIVDLLFKTNRKVTVKQLKEDYFKKIECFD
SVEISGVEDRFNASLGTYHDLLKIIKDKDFLDNEENEDILEDIVLTLTLFEDREM
IEERLKTYAHLFDDKVMKQLKRRRYTGWGRLSRKLINGIRDKQSGKTILDFL
KSDGFANRNFMQLIHDDSLTFKEDIQKAQVSGQGDSLHEHIANLAGSPAIKK
GILQTVKVVDELVKVMGRHKPENIVIEMARENQTTQKGQKNSRERMKRIEEGIK
ELGSQILKEHPVENTQLQNEKLYLYYLQNGRDMYVDQELDINRLSDYDVDAI
VPQSFLKDDSIDNKVLTRSDKNRGKSDNVPSEEVVKKMKNYWRQLLNAKLIT
QRKFDNLTKAERGGLSELDKAGFIKRQLVETRQITKHVAQILDSRMNTKYDEND
KLIREVKVITLKSKLVSDFRKDFQFYKVREINNYHHAHDAYLNAVVGTALIK
KYPKLESEFVYGDYKVYDVRKMIAKSEQEIGKATAKYFFYSNIMNFFKTEI
TLANGEIRKRPLIETNGETGEIVWDKGRDFATVRKVLSMPQVNIVKKTEVQTGGF
SKESILPKRNSDKLIARKKDWDPKKYGGFDSPTVAYSVLVVAKVEKGKSKK
LKSVKELLGITIMERSSFEKNPIDFLEAKGYKEVKKDLIIKLPKYSLFEL
ENGRKRMLASAGELQKGNELALPSKYVNFLYLASHYEKLKGSPEDNEQKQLFVEQ
HKHYLDEIIEQISEFSKRVILADANLDKVLSAYNKHRDKPIREQAENIIHL
FTLTNLGAPAAFKYFDTTIDRKQYRSTKEVLDATLIHQSITGLYETRIDLS
QLGGD
""".replace("\n", "").replace(" ", "")


def test1():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import filter_proteins
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    cas9 = filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    export = to_alphalink2(
        cas9, fasta="data/_fasta/Cas9_plus10.fasta", filename_prefix="Cas9"
    )
    assert export["Exported files"][0] == "Cas9_AlphaLink2.txt"
    assert export["Exported files"][1] == "Cas9_AlphaLink2.fasta"


def test2():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import filter_proteins
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    cas9 = filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    export = to_alphalink2(cas9, fasta="data/_fasta/Cas9_plus10.fasta")
    assert len(
        [
            line
            for line in export["AlphaLink2 crosslinks"].split("\n")
            if line.strip() != ""
        ]
    ) == len(cas9)
    assert export["AlphaLink2 FASTA"].split("\n")[0].split(">")[1].split("|")[0] == "A"
    assert export["AlphaLink2 FASTA"].split("\n")[1].strip() == CAS9
    assert export["AlphaLink2 DataFrame"].shape == (223, 5)
    for i, row in export["AlphaLink2 DataFrame"].iterrows():
        assert (
            int(row["residueFrom"]) == cas9[i]["alpha_proteins_crosslink_positions"][0]
        )
        assert str(row["chain1"]) == "A"
        assert int(row["residueTo"]) == cas9[i]["beta_proteins_crosslink_positions"][0]
        assert str(row["chain2"]) == "A"
        assert float(row["FDR"]) == pytest.approx(0.01)
    assert len(export["Exported files"]) == 0


def test3():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import filter_proteins
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    cas9 = filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    export = to_alphalink2(
        cas9, fasta="data/_fasta/Cas9_plus10.fasta", annotated_fdr=0.02
    )
    assert len(
        [
            line
            for line in export["AlphaLink2 crosslinks"].split("\n")
            if line.strip() != ""
        ]
    ) == len(cas9)
    assert export["AlphaLink2 FASTA"].split("\n")[0].split(">")[1].split("|")[0] == "A"
    assert export["AlphaLink2 FASTA"].split("\n")[1].strip() == CAS9
    assert export["AlphaLink2 DataFrame"].shape == (223, 5)
    for i, row in export["AlphaLink2 DataFrame"].iterrows():
        assert (
            int(row["residueFrom"]) == cas9[i]["alpha_proteins_crosslink_positions"][0]
        )
        assert str(row["chain1"]) == "A"
        assert int(row["residueTo"]) == cas9[i]["beta_proteins_crosslink_positions"][0]
        assert str(row["chain2"]) == "A"
        assert float(row["FDR"]) == pytest.approx(0.02)
    assert len(export["Exported files"]) == 0


def test4():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import filter_proteins
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    cas9 = filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    fdr = [r / len(cas9) for r in range(1, len(cas9) + 1)]
    export = to_alphalink2(
        cas9, fasta="data/_fasta/Cas9_plus10.fasta", annotated_fdr=fdr
    )
    assert len(
        [
            line
            for line in export["AlphaLink2 crosslinks"].split("\n")
            if line.strip() != ""
        ]
    ) == len(cas9)
    assert export["AlphaLink2 FASTA"].split("\n")[0].split(">")[1].split("|")[0] == "A"
    assert export["AlphaLink2 FASTA"].split("\n")[1].strip() == CAS9
    assert export["AlphaLink2 DataFrame"].shape == (223, 5)
    for i, row in export["AlphaLink2 DataFrame"].iterrows():
        assert (
            int(row["residueFrom"]) == cas9[i]["alpha_proteins_crosslink_positions"][0]
        )
        assert str(row["chain1"]) == "A"
        assert int(row["residueTo"]) == cas9[i]["beta_proteins_crosslink_positions"][0]
        assert str(row["chain2"]) == "A"
        assert float(row["FDR"]) == pytest.approx(fdr[i])
    assert len(export["Exported files"]) == 0


def test5():
    from pyXLMS.parser import read
    from pyXLMS import transform
    from pyXLMS.exporter import to_alphalink2

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    pr = transform.annotate_fdr(pr)
    pr = transform.unique(pr, by="protein")
    pr = transform.validate(pr)
    pr = transform.targets_only(pr)
    cas9 = transform.filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]

    fdr = [r / len(cas9) for r in range(1, len(cas9) + 1)]
    annotated_fdr = [
        xl["additional_information"]["pyXLMS_annotated_FDR"] for xl in cas9
    ]
    export = to_alphalink2(
        cas9, fasta="data/_fasta/Cas9_plus10.fasta", annotated_fdr=fdr
    )
    assert len(
        [
            line
            for line in export["AlphaLink2 crosslinks"].split("\n")
            if line.strip() != ""
        ]
    ) == len(cas9)
    assert export["AlphaLink2 FASTA"].split("\n")[0].split(">")[1].split("|")[0] == "A"
    assert export["AlphaLink2 FASTA"].split("\n")[1].strip() == CAS9
    assert export["AlphaLink2 DataFrame"].shape == (223, 5)
    for i, row in export["AlphaLink2 DataFrame"].iterrows():
        assert (
            int(row["residueFrom"]) == cas9[i]["alpha_proteins_crosslink_positions"][0]
        )
        assert str(row["chain1"]) == "A"
        assert int(row["residueTo"]) == cas9[i]["beta_proteins_crosslink_positions"][0]
        assert str(row["chain2"]) == "A"
        assert float(row["FDR"]) == pytest.approx(annotated_fdr[i])
    assert len(export["Exported files"]) == 0


def test6():
    from pyXLMS.parser import read
    from pyXLMS import transform
    from pyXLMS.exporter import to_alphalink2

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    pr = transform.annotate_fdr(pr)
    pr = transform.unique(pr, by="protein")
    pr = transform.validate(pr)
    pr = transform.targets_only(pr)
    cas9 = transform.filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]

    fdr = [r / len(cas9) for r in range(1, len(cas9) + 1)]
    export = to_alphalink2(
        cas9,
        fasta="data/_fasta/Cas9_plus10.fasta",
        annotated_fdr=fdr,
        try_use_annotated_fdr=False,
    )
    assert len(
        [
            line
            for line in export["AlphaLink2 crosslinks"].split("\n")
            if line.strip() != ""
        ]
    ) == len(cas9)
    assert export["AlphaLink2 FASTA"].split("\n")[0].split(">")[1].split("|")[0] == "A"
    assert export["AlphaLink2 FASTA"].split("\n")[1].strip() == CAS9
    assert export["AlphaLink2 DataFrame"].shape == (223, 5)
    for i, row in export["AlphaLink2 DataFrame"].iterrows():
        assert (
            int(row["residueFrom"]) == cas9[i]["alpha_proteins_crosslink_positions"][0]
        )
        assert str(row["chain1"]) == "A"
        assert int(row["residueTo"]) == cas9[i]["beta_proteins_crosslink_positions"][0]
        assert str(row["chain2"]) == "A"
        assert float(row["FDR"]) == pytest.approx(fdr[i])
    assert len(export["Exported files"]) == 0


def test7():
    from pyXLMS.pipelines import pipeline
    from pyXLMS import transform
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    cas9 = transform.filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    fdr = [r / len(cas9) for r in range(1, len(cas9) + 1)][: len(cas9) - 10]
    with pytest.raises(
        ValueError,
        match=(
            "Length of annotated_fdr does not match length of crosslinks! "
            "When providing a list it needs to contain FDR values for every crosslink and therefore be of equal length!"
        ),
    ):
        _export = to_alphalink2(
            cas9, fasta="data/_fasta/Cas9_plus10.fasta", annotated_fdr=fdr
        )


def test8():
    from pyXLMS.pipelines import pipeline
    from pyXLMS import transform
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    cas9 = transform.filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    with pytest.raises(
        IndexError,
        match="Found more than the supported 62 proteins/chains in the fasta file! Please trim fasta file to a maximum of 62 sequences!",
    ):
        _export = to_alphalink2(
            cas9,
            fasta="data/_fasta/Ribosome_for_XL_all_contained_Proteins_from_shotgun.fasta",
        )


def test9():
    from pyXLMS.exporter.to_alphalink2 import __get_proteins_and_positions

    proteins_and_positions = __get_proteins_and_positions(
        "GSQKDR", {"A": {"header": "Cas9", "sequence": CAS9}}
    )
    assert len(proteins_and_positions[0]) == 0
    assert len(proteins_and_positions[1]) == 0


def test10():
    from pyXLMS.exporter.to_alphalink2 import __get_proteins_and_positions

    with pytest.raises(RuntimeError, match="No match found for peptide GSQKDR!"):
        _proteins_and_positions = __get_proteins_and_positions(
            "GSQKDR", {"A": {"header": "Cas9", "sequence": CAS9}}, True
        )


def test11():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    export = to_alphalink2(pr["crosslinks"], fasta="data/_fasta/Cas9.fasta", verbose=0)
    assert export["AlphaLink2 DataFrame"].shape[0] == 223


def test12():
    from pyXLMS.parser import read
    from pyXLMS.transform import targets_only
    from pyXLMS.transform import filter_proteins
    from pyXLMS.exporter import to_alphalink2

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    should = len(
        filter_proteins(targets_only(pr["crosslinks"]), proteins=["Cas9"])["Both"]
    )
    assert len(pr["crosslinks"]) == 300
    export = to_alphalink2(pr["crosslinks"], fasta="data/_fasta/Cas9.fasta", verbose=0)
    assert export["AlphaLink2 DataFrame"].shape[0] == should


def test13():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.warns(
        RuntimeWarning,
        match=r"Could not find matching proteins in FASTA file for crosslink id 148:SDNVPSEEVVKKMK-VASMASEKMK! This warning can be ignored if this is to be expected\.",
    ):
        _export = to_alphalink2(pr["crosslinks"], fasta="data/_fasta/Cas9.fasta")


def test14():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.exporter import to_alphalink2

    pr = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
        engine="MS Annika",
        crosslinker="DSS",
    )
    with pytest.raises(
        RuntimeError,
        match=r"Could not find matching proteins in FASTA file for crosslink id 148:SDNVPSEEVVKKMK-VASMASEKMK! If this is to be expected please set verbose level to either 1 or 0!",
    ):
        _export = to_alphalink2(
            pr["crosslinks"], fasta="data/_fasta/Cas9.fasta", verbose=2
        )
